use std::{env, path::Path};

use argon2::{
    password_hash::{rand_core::OsRng, SaltString},
    Argon2, PasswordHasher,
};
use axum::{
    http::StatusCode,
    response::{Html, IntoResponse},
    routing::get,
    Form, Router,
};
use axum_csrf::{CsrfConfig, CsrfLayer, CsrfToken};
use maud::html;
use notify::Watcher;
use serde::Deserialize;
use tower_http::services::ServeDir;
use tower_livereload::LiveReloadLayer;
use tracing::instrument;

use base64::prelude::*;

use crate::{error::AppError, models::NewUser, state::State, views};

#[instrument]
pub async fn new() -> Result<Router, anyhow::Error> {
    let app_router = Router::new()
        .route("/hello-world", get(views::greet_world))
        .route("/hello-world-text", get(views::greet_world_text));

    let assets_dir = ServeDir::new("./assets");
    let state = State::new().await?;

    let live_reload_layer = LiveReloadLayer::new();
    let reloader = live_reload_layer.reloader();

    let mut watcher = notify::recommended_watcher(move |_| reloader.reload())?;
    watcher.watch(Path::new("assets"), notify::RecursiveMode::Recursive)?;

    let cookie_key_bytes: Vec<u8> = BASE64_URL_SAFE_NO_PAD
        .decode(env::var("COOKIE_KEY").expect("COOKIE_KEY not set"))
        .expect("COOKIE_KEY not base64 URL-safe encoded");
    let cookie_key = cookie::Key::from(&cookie_key_bytes);
    let csrf_config = CsrfConfig::default().with_key(Some(cookie_key));

    let router = Router::new()
        .fallback(|| async { (StatusCode::NOT_FOUND, "404 page not found") })
        .nest("/app", app_router)
        .nest_service("/assets", assets_dir)
        .route("/", get(views::index))
        .route("/login", get(views::login).post(login))
        .route("/register", get(views::register).post(register))
        .route("/all_users", get(views::all_users))
        .with_state(state)
        .layer(CsrfLayer::new(csrf_config))
        .layer(live_reload_layer);

    Ok(router)
}

#[derive(Deserialize)]
struct Register {
    authenticity_token: String,
    username: String,
    password: String,
}

impl<'a> TryInto<NewUser<'a>> for &'a Register {
    type Error = anyhow::Error;

    fn try_into(self: &'a Register) -> Result<NewUser<'a>, Self::Error> {
        let salt = SaltString::generate(&mut OsRng);
        let argon2 = Argon2::default();
        let password_digest: Vec<u8> = argon2
            .hash_password(self.password.as_bytes(), &salt)?
            .hash
            .expect("no password hash")
            .as_bytes()
            .into();
        Ok(NewUser {
            username: &self.username,
            name: None,
            password_digest,
        })
    }
}

async fn register(
    c: CsrfToken,
    Form(register): Form<Register>,
) -> Result<impl IntoResponse, AppError> {
    // TODO: https://docs.rs/axum_csrf/latest/axum_csrf/#prevent-post-replay-attacks-with-csrf
    c.verify(&register.authenticity_token)?;

    let new_user: NewUser = (&register).try_into()?;

    Ok((
        StatusCode::CREATED,
        Html(
            html! {
                h1 { (new_user.username) }
            }
            .into_string(),
        ),
    ))
}

#[derive(Deserialize)]
struct Login {
    authenticity_token: String,
    username: String,
    password: String,
}

async fn login(
    csrf_token: CsrfToken,
    Form(register): Form<Login>,
) -> Result<impl IntoResponse, AppError> {
    // TODO: https://docs.rs/axum_csrf/latest/axum_csrf/#prevent-post-replay-attacks-with-csrf

    let v = csrf_token.verify(&register.authenticity_token);
    println!("{:?} {:?}", register.authenticity_token, v);
    if v.is_err() {
        return Ok((
            StatusCode::BAD_REQUEST,
            Html(html! { "invalid request" }.into_string()),
        ));
    }

    Ok((
        StatusCode::CREATED,
        Html(
            html! {
                h1 { (register.username) }
                h1 { (register.password) }
            }
            .into_string(),
        ),
    ))
}
