use axum::{
    extract::State,
    response::{Html, IntoResponse},
};
use axum_csrf::CsrfToken;
use maud::html;
use tracing::instrument;

use crate::{
    error::AppError,
    models::User,
    partials::{footer, header},
};

pub async fn csrf<F>(csrf: CsrfToken, cb: F) -> impl IntoResponse
where
    F: Fn(&str) -> Html<String>,
{
    let token = csrf.authenticity_token().unwrap();
    (csrf, cb(&token))
}

#[instrument]
pub async fn index() -> Html<String> {
    Html(
        html! {
            (header())
            main class="prose" {
                h1 { "Manage live lyrics and music displays" }
                p { "Stop struggling to share the same messy set of PowerPoint files or Google Presentations. Make editing and controlling your live lyrics and music displays easy and simple." }
                ul {
                    li { "Live collaboration with your team" }
                    li { "Fully compatible with any device" }
                    li { "Simple workflow" }
                    li { "Dark theme and light theme" }
                    li { "Generous free plan" }
                    li { "Lightweight and fast" }
                }
                section class="flex gap" {
                    a href="/register" class="button bg-primary" { "Try now" }
                    a class="button" href="/login" { "Login" }
                }
            }
            (footer())
        }
        .into_string(),
    )
}

pub async fn register(t: CsrfToken) -> impl IntoResponse {
    csrf(t, |token| {
        Html(
            html! {
                (header())
                main class="prose" {
                    h1 { "Register an account" }
                    form method="post" {
                        input type="hidden" name="authenticity_token" value=(token) {}
                        label {
                            "Username:"
                            input name="username" {}
                        }
                        label {
                            "Password:"
                            input type="password" name="password" {}
                        }
                        button type="submit" { "Create Account" }
                    }
                }
                (footer())
            }
            .into_string(),
        )
    })
    .await
}

pub async fn login(t: CsrfToken) -> impl IntoResponse {
    csrf(t, |token| {
        Html(
            html! {
                (header())
                main class="prose" {
                    h1 { "Login" }
                    form method="post" {
                        input type="hidden" name="authenticity_token" value=(token) {}
                        label {
                            input {}
                        }
                        label {
                            input {}
                        }
                    }
                }
                (footer())
            }
            .into_string(),
        )
    })
    .await
}

#[allow(unreachable_code)]
pub async fn all_users(
    State(_state): State<crate::state::State>,
) -> Result<Html<String>, AppError> {
    let all_users: Vec<User> = vec![];

    // @if let Some(name) = u.name {
    //     name
    // } @else {
    //     "N/A"
    // }
    Ok(Html(
        html! {
            (header())
            main class="prose" {
                h1 { "Users" }
                ul {
                    @for u in all_users {
                        li {
                            (u.username)
                            @if let Some(name) = u.name {
                                " ("
                                (name)
                                ")"
                            }
                        }
                    }
                }
            }
            (footer())
        }
        .into_string(),
    ))
}

#[instrument]
pub async fn greet_world() -> Html<String> {
    Html(
        html! {
            h1 { (greet_world_text().await) }
        }
        .into_string(),
    )
}

#[instrument]
pub async fn greet_world_text() -> &'static str {
    "Hello, World!"
}
